{-# LANGUAGE ImportQualifiedPost #-}
module Language.Brainfuck.Instruction.Compressed ( CompressedInstruction(..), compress ) where

import Data.Bifunctor ( Bifunctor(first) )
import Data.Vector ( Vector )
import Data.Word ( Word8 )
import Numeric.Natural ( Natural )

import Data.Vector qualified as Vector

import Language.Brainfuck.Instruction ( Instruction )
import Language.Brainfuck.Instruction qualified as Instruction 

data CompressedInstruction
  = Add Word8
  | Subtract Word8
  | MoveRight Natural
  | MoveLeft Natural
  | ReadByte
  | PutByte
  | Loop (Vector CompressedInstruction)
  deriving (Show, Eq)

compress :: Vector Instruction -> Vector CompressedInstruction
compress instructions = Vector.fromList (go instructions)
  where
    go program = case Vector.uncons program of
      Nothing -> []
      Just (x, rest) -> let

        withRepeatCount f = let (count, rest') = spanLength x rest
          in (f . fromIntegral . succ) count : go rest'
          
        in case x of
          Instruction.Increment  -> withRepeatCount Add
          Instruction.Decrement  -> withRepeatCount Subtract
          Instruction.MoveRight  -> withRepeatCount MoveRight
          Instruction.MoveLeft   -> withRepeatCount MoveLeft
          Instruction.ReadByte   -> ReadByte : go rest
          Instruction.PutByte    -> PutByte : go rest
          Instruction.Loop inner -> Loop (compress inner) : go rest

    spanLength :: Eq a => a -> Vector a -> (Int, Vector a)
    spanLength x xs = first Vector.length $ Vector.span (== x) xs

-- >>> fromInteger 300 :: Word8
-- 44

-- >>> compress . Vector.fromList $ replicate 300 Instruction.Increment
-- [Add 44]
-- >>> compress . Vector.fromList $ replicate 300 Instruction.MoveRight
-- [MoveRight 300]

